/**********************************************************************
audio play & record (Linux, Solaris, IRIX)

$Id: playrec.c,v 1.11 1999/12/28 15:23:50 purnhage Exp $

Header file: playrec.h

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>
NM    Nikolaus Meine, Uni Hannover (c/o Heiko Purnhagen)
BE    Bernd Edler, Uni Hannover <edler@tnt.uni-hannover.de>

Changes:
16-sep-98   HP/NM   born, based on auplay.c and aurec.c
            HP/BE   ... and play_list.c
17-sep-98   HP      Solaris & IRIX
07-jan-99   HP      warning if no audio device support
08-jan-99   HP      audio device selection, fixed Solaris PROpenRead()
26-jan-99   HP      SGI: AL_DIGITAL_INPUT_RATE
18-feb-99   HP      write 0 before closing play device
23-feb-99   HP      fill up to n*numChan before closing play device
                    (prevents lock on Linux)
11-nov-99   HP      fixed debug output in PROpenWrite: device=
14-nov-99   HP      removed fileno(), using open() instead fopen()
19-nov-99   HP      added PRWriteBuffer(), PRWriteFilled(), PRNap()
21-nov-99   HP      fixed bug in PRNap() / work-around for alsa-0.3.0 startup
27-nov-99   HP      fixed PRWriteFilled for old OSS/ES1688 (2.0.35)
28-dec-99   HP      fixed PRReadBuffer() return value (Linux/Solaris)
**********************************************************************/

/* Audio play & record for Solaris, Linux, IRIX. */
/* Sample format: 16 bit twos complement, uniform quantisation */

/* Multi channel data is interleaved: l0 r0 l1 r1 ... */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "playrec.h"
#include "common.h"

#ifdef Linux
#include <sys/ioctl.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#endif

#ifdef Solaris
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/audioio.h>
#endif

#ifdef IRIX
#include <unistd.h>
#include <dmedia/audio.h>
#endif


/* ---------- declarations ---------- */

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define IRIXSIZE (2.0)		/* IRIX audio buffer size (sec) */


/* ---------- declarations (structures) ---------- */

struct PRStreamStruct		/* play/rec stream handle */
{
#ifdef Linux
  int fd;
#endif

#ifdef Solaris
  int fd;
#endif

#ifdef IRIX
  ALconfig rt_config;
  ALport rt_audio_p;
#endif

  long currentSample;		/* number of samples read/written */
  int write;			/* write (not read) flag */
  int numChannel;		/* number of channels */
  float fSample;		/* sampling frequency [Hz] */
  long bufNumSample;		/* play buffer size */
};


/* ---------- variables ---------- */

static int PRdebugLevel = 0;	/* debug level */


/* ---------- functions ---------- */


/* PRInit() */
/* Init audio play & record. */

void PRInit (
  int debugLevel)		/* in: debug level */
				/*     0=off  1=basic  2=full */
{
  PRdebugLevel = debugLevel;
  if (PRdebugLevel >= 1) {
    printf("PRInit: debugLevel=%d\n",PRdebugLevel);
  }
}


/* PRNap() */
/* nap for ms milliseconds */

void PRNap (
  int ms)			/* in: time to nap [ms] */
{
#ifdef Linux
  struct timeval tv;

  tv.tv_sec = 0;
  tv.tv_usec = min(999999,max(1,ms)*1000);
  select(0,NULL,NULL,NULL,&tv);
#endif  

#ifdef Solaris
  struct timeval tv;

  tv.tv_sec = 0;
  tv.tv_usec = min(999999,max(1,ms)*1000);
  select(0,NULL,NULL,NULL,&tv);
#endif  

#ifdef IRIX
  sginap(max(1,ms/CLK_TCK));
#endif
}


/* PROpenRead() */
/* Open play/rec stream for reading. */

PRStream *PROpenRead (
  char *device,			/* in: device name (NULL = default device) */
  int *numChannel,		/* in/out: number of channels */
  float *fSample)		/* in/out: sampling frequency [Hz] */
				/*  numChannel & fSample: */
				/*  in: requested configuration */
				/*      (SGI: 0 = use audio panel settings) */
				/*  out: selected configuration */
				/*       (set according to audio */
				/*        device capabilities) */
				/* returns: */
				/*  play/rec stream (handle) */
				/*  or NULL if error */
{
  PRStream *s;

#ifdef Linux
  int err,tmp;
#endif  

#ifdef Solaris
  audio_info_t ainfo;
  int err;
#endif  

#ifdef IRIX
  long sr[2];
  long size;
  int err;
#endif  
  
  if (PRdebugLevel)
    printf("PROpenRead: device=\"%s\" numChannel=%d fSample=%.1f\n",
	   device?device:"(default)",*numChannel,*fSample);

  if ((s=(PRStream*)malloc(sizeof(PRStream)))==NULL)
    CommonExit(-1,"PROpenRead: Can not allocate memory");
  s->currentSample = 0;
  s->write = 0;

#ifdef Linux
  if (!device)
    device = "/dev/dsp";
  if ((s->fd=open(device,O_RDONLY,0))==-1) {
    CommonWarning("PROpenRead: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  tmp = AFMT_S16_LE;			/* aka 16 */
  err = ioctl(s->fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != AFMT_S16_LE) {
    CommonWarning("PROpenRead: 16 bit audio not supported");
    free(s);
    return NULL;
  }
  tmp = (*numChannel>1)?1:0;
  err = err || ioctl(s->fd, SNDCTL_DSP_STEREO, &tmp);
  *numChannel = tmp+1;
  tmp = (int)(*fSample+.5);
  err = err || ioctl (s->fd, SNDCTL_DSP_SPEED, &tmp);
  *fSample = tmp;
  if (err) {
    CommonWarning("PROpenRead: Unable to init dsp");
    free(s);
    return NULL;
  }
#endif

#ifdef Solaris
  if (!device)
    device = "/dev/audio";
  if ((s->fd=open(device,O_RDONLY,0))==-1) {
    CommonWarning("PROpenRead: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  AUDIO_INITINFO(&ainfo);

  ainfo.record.sample_rate = (long)(*fSample+.5);
  ainfo.record.channels = *numChannel;
  ainfo.record.precision	= 16;
  ainfo.record.encoding	= AUDIO_ENCODING_LINEAR;
  
  err = ioctl(s->fd,AUDIO_SETINFO,&ainfo);
  err = err || ioctl(s->fd,AUDIO_GETINFO,&ainfo);
  *fSample = ainfo.record.sample_rate;
  *numChannel = ainfo.record.channels;
  if (err) {
    CommonWarning("PROpenRead: Unable to init audio");
    free(s);
    return NULL;
  }
  if (ainfo.record.precision != 16 ||
      ainfo.record.encoding != AUDIO_ENCODING_LINEAR) {
    CommonWarning("PROpenRead: 16 bit audio not supported");
    free(s);
    return NULL;
  }
#endif

#ifdef IRIX
  sr[0] = AL_INPUT_RATE;
  sr[1] = (long)(*fSample+.5);
  err = 0;
  if (*fSample>0)
    err = err || ALsetparams(AL_DEFAULT_DEVICE,sr,2);
  err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
  *fSample = sr[1];
  if (*fSample == AL_RATE_AES_1) {
    sr[0] = AL_DIGITAL_INPUT_RATE;
    err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
    *fSample = sr[1];
  }
  s->rt_config = ALnewconfig();
  err = err || (!s->rt_config);
  err = err || ALsetsampfmt(s->rt_config,AL_SAMPFMT_TWOSCOMP);
  err = err || ALsetwidth(s->rt_config,AL_SAMPLE_16); 
  if (*numChannel>0)
    err = err || ALsetchannels(s->rt_config,*numChannel);
  err = err || ((*numChannel=ALgetchannels(s->rt_config))<0);
  size = (long)(*numChannel**fSample*IRIXSIZE);
  if (PRdebugLevel)
    printf("PROpenRead: got: numChannel=%d fSample=%.1f size=%ld\n",
	   *numChannel,*fSample,size);
  err = err || ALsetqueuesize(s->rt_config,size);
  if (err || !(s->rt_audio_p = ALopenport("adc","r",s->rt_config))) {
    CommonWarning("PROpenRead: Unable to init adc");
    free(s);
    return NULL;
  }
#endif

#if !(defined Linux || defined Solaris || defined IRIX)
  CommonWarning("PROpenRead: no audio device support");
  free(s);
  return NULL;
#endif

  s->numChannel = *numChannel;
  s->fSample = *fSample;

  if (PRdebugLevel)
    printf("PROpenRead: set to: numChannel=%d fSample=%.1f\n",
	   *numChannel,*fSample);

  return s;
}


/* PROpenWrite() */
/* Open play/rec stream for writing. */

PRStream *PROpenWrite (
  char *device,			/* in: device name (NULL = default device) */
  int *numChannel,		/* in/out: number of channels */
  float *fSample)		/* in/out: sampling frequency [Hz] */
				/*  numChannel & fSample: */
				/*  in: requested configuration */
				/*      (SGI: 0 = use audio panel settings) */
				/*  out: selected configuration */
				/*       (set according to audio */
				/*        device capabilities) */
				/* returns: */
				/*  play/rec stream (handle) */
				/*  or NULL if error */
{
  PRStream *s;

#ifdef Linux
  int err,tmp;
  audio_buf_info ainfo;
#endif  

#ifdef Solaris
  audio_info_t ainfo;
  int err;
#endif  

#ifdef IRIX
  long sr[2];
  long size;
  int err;
#endif  
  
  if (PRdebugLevel)
    printf("PROpenWrite: device=\"%s\" numChannel=%d fSample=%.1f\n",
	   device?device:"(default)",*numChannel,*fSample);

  if ((s=(PRStream*)malloc(sizeof(PRStream)))==NULL)
    CommonExit(-1,"PROpenWrite: Can not allocate memory");
  s->currentSample = 0;
  s->write = 1;
  s->bufNumSample = 0;

#ifdef Linux
  if (!device)
    device = "/dev/dsp";
  if ((s->fd=open(device,O_WRONLY,0))==-1) {
    CommonWarning("PROpenWrite: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  /* use 20 ms .. 40 ms fragement size */
  tmp = 9;				/* min 512 bytes/fragment */
  while (1<<tmp < (int)(0.020*(*fSample)+.5)*2*(*numChannel))
    tmp++;
  tmp |= 0x7fff0000;
  err = ioctl(s->fd, SNDCTL_DSP_SETFRAGMENT, &tmp);

  tmp = AFMT_S16_LE;			/* aka 16 */
  err = err || ioctl(s->fd, SNDCTL_DSP_SAMPLESIZE, &tmp);
  if (tmp != AFMT_S16_LE) {
    CommonWarning("PROpenWrite: 16 bit audio not supported");
    free(s);
    return NULL;
  }
  tmp = (*numChannel>1)?1:0;
  err = err || ioctl(s->fd, SNDCTL_DSP_STEREO, &tmp);
  *numChannel = tmp+1;
  tmp = (int)(*fSample+.5);
  err = err || ioctl (s->fd, SNDCTL_DSP_SPEED, &tmp);
  *fSample = tmp;
  if (err) {
    CommonWarning("PROpenWrite: Unable to init dsp");
    free(s);
    return NULL;
  }

  ioctl(s->fd,SNDCTL_DSP_GETOSPACE,&ainfo);

  if (PRdebugLevel)
    printf("PROpenWrite: fragstotal=%d fragsize=%d\n",
	   ainfo.fragstotal,ainfo.fragsize);
#endif

#ifdef Solaris
  if (!device)
    device = "/dev/audio";
  if ((s->fd=open(device,O_WRONLY,0))==-1) {
    CommonWarning("PROpenWrite: Can not open \"%s\"",device);
    free(s);
    return NULL;
  }

  AUDIO_INITINFO(&ainfo);

  ainfo.play.sample_rate = (long)(*fSample+.5);
  ainfo.play.channels = *numChannel;
  ainfo.play.precision	= 16;
  ainfo.play.encoding	= AUDIO_ENCODING_LINEAR;
  
  err = ioctl(s->fd,AUDIO_SETINFO,&ainfo);
  err = err || ioctl(s->fd,AUDIO_GETINFO,&ainfo);
  *fSample = ainfo.play.sample_rate;
  *numChannel = ainfo.play.channels;
  if (err) {
    CommonWarning("PROpenWrite: Unable to init audio");
    free(s);
    return NULL;
  }
  if (ainfo.play.precision != 16 ||
      ainfo.play.encoding != AUDIO_ENCODING_LINEAR) {
    CommonWarning("PROpenWrite: 16 bit audio not supported");
    free(s);
    return NULL;
  }

  if (PRdebugLevel)
    printf("PROpenWrite: buffer_size=%d\n",ainfo.play.buffer_size);
#endif

#ifdef IRIX
  sr[0]=AL_OUTPUT_RATE;
  sr[1]=(long)(*fSample+.5);
  err = 0;
  if (*fSample>0)
    err = err || ALsetparams(AL_DEFAULT_DEVICE,sr,2);
  err = err || ALgetparams(AL_DEFAULT_DEVICE,sr,2);
  *fSample = sr[1];
  s->rt_config = ALnewconfig();
  err = err || (!s->rt_config);
  err = err || ALsetsampfmt(s->rt_config,AL_SAMPFMT_TWOSCOMP);
  err = err || ALsetwidth(s->rt_config,AL_SAMPLE_16); 
  if (*numChannel>0)
    err = err || ALsetchannels(s->rt_config,*numChannel);
  err = err || ((*numChannel=ALgetchannels(s->rt_config))<0);
  size = (long)(*numChannel**fSample*IRIXSIZE);
  if (PRdebugLevel)
    printf("PROpenWrite: size=%ld\n",size);
  err = err || ALsetqueuesize(s->rt_config,size);
  if (err || !(s->rt_audio_p = ALopenport("dac","w",s->rt_config))) {
    CommonWarning("PROpenWrite: Unable to init dac");
    free(s);
    return NULL;
  }
#endif

#if !(defined Linux || defined Solaris || defined IRIX)
  CommonWarning("PROpenWrite: no audio device support");
  free(s);
  return NULL;
#endif

  s->numChannel = *numChannel;
  s->fSample = *fSample;

  if (PRdebugLevel)
    printf("PROpenWrite: set to: numChannel=%d fSample=%.1f\n",
	   *numChannel,*fSample);

  return s;
}


/* PRReadData() */
/* Read data from play/rec stream. */

long PRReadData (
  PRStream *stream,		/* in: play/rec stream (handle) */
  short *data,			/* out: data[] */
  long numSample)		/* in: number of samples to be read */
				/* returns: */
				/*  number of samples read */
{
  long i;

  if (PRdebugLevel > 1)
    printf("PRReadData: numSample=%ld\n",numSample);

  if (stream->write)
    CommonExit(1,"PRReadData: device not in read mode");

#ifdef Linux
  i = read(stream->fd,data,sizeof(short)*numSample);
  if (i>0)
    i /= sizeof(short);
#endif

#ifdef Solaris
  i = read(stream->fd,data,sizeof(short)*numSample);
  if (i>0)
    i /= sizeof(short);
#endif

#ifdef IRIX
  i = (ALreadsamps(stream->rt_audio_p,data,numSample))?0:numSample;
#endif

  stream->currentSample += i;
  return i;
}


/* PRWriteBuffer() */
/* Set play buffer size. */
/* PRWriteData() waits until less than bufNumSamples are left in play */
/* buffer before writing the new samples. */

void PRWriteBuffer (
  PRStream *stream,		/* in: play/rec stream (handle) */
  long bufNumSample)		/* in: number of samples in play buffer */
				/*     (0 to use max. available buffer) */
{
  if (PRdebugLevel > 1)
    printf("PRWriteBuffer: bufNumSample=%ld\n",bufNumSample);

  if (!stream->write)
    CommonExit(1,"PRWriteBuffer: device not in write mode");

  stream->bufNumSample = bufNumSample;
}


/* PRWriteFilled() */
/* Get number of samples currently left in play buffer, */
/* i.e. written by PRWriteData() but not yet played. */

long PRWriteFilled (
  PRStream *stream)		/* in: play/rec stream (handle) */
				/* returns: */
				/*  number of samples left in play buffer */
{
  long numSampleLeft;

#ifdef Linux
  count_info cinfo;
  audio_buf_info ainfo;
  long tmp;
#endif

#ifdef Solaris
  audio_info_t ainfo;
#endif

  if (!stream->write)
    CommonExit(1,"PRWriteFilled: device not in write mode");

#ifdef Linux
  ioctl(stream->fd,SNDCTL_DSP_GETOPTR,&cinfo);
  ioctl(stream->fd,SNDCTL_DSP_GETOSPACE,&ainfo);
  numSampleLeft = stream->currentSample-cinfo.bytes/2;
  /* work-around for old OSS/ES1688 (2.0.35) -- (fragLeft-1)*fragSize */
  tmp = ((ainfo.fragstotal-1)*ainfo.fragsize-ainfo.bytes)/2;
  if (PRdebugLevel > 1)
    printf("bts=%d bks=%d ptr=%d "
	   "fgm=%d (%d*%d) bts=%d "
	   "cur=%ld lft=%ld/%ld\n",
	   cinfo.bytes,cinfo.blocks,cinfo.ptr,ainfo.fragments,
	   ainfo.fragstotal,ainfo.fragsize,ainfo.bytes,
	   stream->currentSample,numSampleLeft,tmp);
  numSampleLeft=max(0,numSampleLeft);
  tmp = max(0,tmp);
  numSampleLeft = max(numSampleLeft,tmp);
  /* make sure that play starts ... */
  if (numSampleLeft < ainfo.fragsize/2)
    numSampleLeft = 0;
#endif

#ifdef Solaris
  ioctl(stream->fd,AUDIO_GETINFO,&ainfo);
  numSampleLeft = stream->currentSample-ainfo.play.samples*stream->numChannel;
#endif

#ifdef IRIX
  numSampleLeft = ALgetfilled(stream->rt_audio_p);
#endif

  if (PRdebugLevel > 1)
    printf("PRWriteFilled: numSampleLeft=%ld\n",numSampleLeft);

  return numSampleLeft;
}


/* PRWriteData() */
/* Write data to play/rec stream. */

void PRWriteData (
  PRStream *stream,		/* in: play/rec stream (handle) */
  short *data,			/* in: data[] */
  long numSample)		/* in: number of samples to be written */
{
  if (PRdebugLevel > 1)
    printf("PRWriteData: numSample=%ld\n",numSample);

  if (!stream->write)
    CommonExit(1,"PRWriteData: device not in write mode");

  if (stream->bufNumSample)
    while (PRWriteFilled(stream)>stream->bufNumSample)
      PRNap(10);				/* 10 ms */

#ifdef Linux
  write(stream->fd,data,sizeof(short)*numSample);

  /* work-around for alsa-0.3.0 -- wait until playback started */
  if (stream->bufNumSample &&
      stream->currentSample <= stream->bufNumSample &&
      stream->currentSample+numSample > stream->bufNumSample) {
    count_info cinfo;
    audio_buf_info ainfo;
    ioctl(stream->fd,SNDCTL_DSP_GETOPTR,&cinfo);
    ioctl(stream->fd,SNDCTL_DSP_GETOSPACE,&ainfo);
    if (numSample*2>=ainfo.fragsize)
      while (cinfo.ptr==0x20000) {
	PRNap(1);
	ioctl(stream->fd,SNDCTL_DSP_GETOPTR,&cinfo);
      }
    if (PRdebugLevel > 1)
      printf("alsa started: bts=%d blks=%d ptr=%d / "
	     "fgmts=%d ftot=%d frgsz=%d bts=%d\n",
	     cinfo.bytes,cinfo.blocks,cinfo.ptr,ainfo.fragments,
	     ainfo.fragstotal,ainfo.fragsize,ainfo.bytes);
  }
#endif

#ifdef Solaris
  write(stream->fd,data,sizeof(short)*numSample);
#endif

#ifdef IRIX
  ALwritesamps(stream->rt_audio_p,data,numSample);
#endif

  stream->currentSample += numSample;
}


/* PRClose() */
/* Close play/rec stream.*/

void PRClose (
  PRStream *stream)		/* in: play/rec stream (handle) */
{
  short data[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  int n;

  if (PRdebugLevel)
    printf("PRClose: currentSample=%ld\n",stream->currentSample);

  n = stream->currentSample%stream->numChannel;
  if (n)
    CommonWarning("PRClose: odd number of samples (smps%%chan=%d chan=%d)",
		  n,stream->numChannel);

  if (stream->write) {
    if (n)
      n = stream->numChannel-n;		/* missing samples */
    n += stream->numChannel;		/* silence */
    n = min(16,n);			/* ugly, but ... */

#ifdef Linux
    write(stream->fd,data,sizeof(short)*n);
#endif

#ifdef Solaris
    write(stream->fd,data,sizeof(short)*n);
#endif

#ifdef IRIX
    ALwritesamps(stream->rt_audio_p,data,n);
#endif
  }

#ifdef Linux
  ioctl(stream->fd, SNDCTL_DSP_SYNC, 0);
  /* ioctl(stream->fd, SNDCTL_DSP_RESET, 0); */
  close(stream->fd);
#endif

#ifdef Solaris
  close(stream->fd);
#endif

#ifdef IRIX
  if (stream->write)
    while (ALgetfilled(stream->rt_audio_p) > 0)
      sginap(1);			/* typ. 10 ms (CLK_TCK = 100) */
  ALcloseport(stream->rt_audio_p);
  ALfreeconfig(stream->rt_config);
#endif

  free(stream);
}


/* end of austream.c */

